@@ -4,33 +4,44 @@
 # Conf files subroutines brought by the the Project-Builder project
 # which can be easily used by wahtever perl project
 #
+# Copyright B. Cornec 2007-2012
+# Eric Anderson's changes are (c) Copyright 2012 Hewlett Packard
+# Provided under the GPL v2
+#
 # $Id$
 #
 
 package ProjectBuilder::Conf;
 
 use strict;
+use Carp 'confess';
 use Data::Dumper;
 use ProjectBuilder::Base;
+use ProjectBuilder::Version;
 
 # Inherit from the "Exporter" module which handles exporting functions.
  
+use vars qw($VERSION $REVISION @ISA @EXPORT);
 use Exporter;
  
 # Export, by default, all the functions into the namespace of
 # any code which uses this module.
  
 our @ISA = qw(Exporter);
-our @EXPORT = qw(pb_conf_init pb_conf_add pb_conf_read pb_conf_read_if pb_conf_get pb_conf_get_if);
+our @EXPORT = qw(pb_conf_init pb_conf_add pb_conf_read pb_conf_read_if pb_conf_get pb_conf_get_if pb_conf_print pb_conf_get_all);
+($VERSION,$REVISION) = pb_version_init();
 
 # Global hash of conf files
 # Key is the conf file name
 # Value is its rank
 my %pbconffiles;
 
-# Global hash of cached values. 
+# Global hash of conf file content
+# Key is the config keyword
+# Value is a hash whose key depends on the nature of the config keyword as documented
+# and value is the confguration value
 # We consider that values can not change during the life of pb
-# my %cachedval;
+my $h = ();
 
 =pod
 
@@ -68,35 +79,86 @@ sub pb_conf_init {
 
 my $proj=shift || undef;
 
+pb_log(1,"Entering pb_conf_init\n");
+#
+# Check project name
+# Could be with env var PBPROJ
+# or option -p
+# if not defined take the first in conf file
+#
+if ((defined $ENV{'PBPROJ'}) &&
+	(not defined $proj)) {
+	pb_log(2,"PBPROJ env var setup ($ENV{'PBPROJ'}) so using it\n");
+	$proj = $ENV{'PBPROJ'};
+}
+
 if (defined $proj) {
 	$ENV{'PBPROJ'} = $proj;
 } else {
 	$ENV{'PBPROJ'} = "default";
 }
+pb_log(1,"PBPROJ = $ENV{'PBPROJ'}\n");
 }
 
 
+=item B<pb_conf_cache>
+
+This function caches the configuration file content passed as first parameter into the a hash passed in second parameter
+It returns the modified hash
+Can be used in correlation with the %h hash to store permanently values or not if temporarily.
+
+=cut
+
+sub pb_conf_cache {
+
+my $cf = shift;
+my $lh = shift;
+
+# Read the content of the config file and cache it in the %h hash further availble for queries
+open(CONF,$cf) || confess "Unable to open $cf";
+while(<CONF>) {
+	if (/^\s*([A-z0-9-_.]+)\s+([[A-z0-9-_.]+)\s*=\s*(.*)$/) {
+		pb_log(3,"DEBUG: 1:$1 2:$2 3:$3\n");
+		$lh->{$1}->{$2}=$3;
+	}
+}
+close(CONF);
+return($lh);
+}
 
 =item B<pb_conf_add>
 
-This function adds the configuration file to the list last.
+This function adds the configuration file to the list last, and cache their content in the %h hash
 
 =cut
 
 sub pb_conf_add {
 
 pb_log(2,"DEBUG: pb_conf_add with ".Dumper(@_)."\n");
+my $lh;
 
 foreach my $cf (@_) {
+	if (! -r $cf) {
+		pb_log(0,"WARNING: pb_conf_add can not read $cf\n");
+		next;
+	}
 	# Skip already used conf files
-	next if (defined $pbconffiles{$cf});
+	return($lh) if (defined $pbconffiles{$cf});
+	
 	# Add the new one at the end
 	my $num = keys %pbconffiles;
-	pb_log(2,"DEBUG: pb_conf_add $cf at position $num\n");
+	pb_log(2,"DEBUG: pb_conf_cache of $cf at position $num\n");
 	$pbconffiles{$cf} = $num;
+
+	# Read the content of the config file 
+	$lh = pb_conf_cache($cf,$lh);
+	# and cache it in the %h hash for further queries but after the previous
+	# as we load conf files in reverse order (most precise first)
+	pb_conf_add_last_in_hash($lh)
 }
 }
 
+
 =item B<pb_conf_read_if>
 
 This function returns a table of pointers on hashes
@@ -126,6 +188,8 @@ will allow to get the mapping:
 
 Valid chars for keys and tags are letters, numbers, '-' and '_'.
 
+The file read is forgotten after its usage. If you want permanent caching of the data, use pb_conf_add then pb_conf_get
+
 =cut
 
 sub pb_conf_read_if {
@@ -148,34 +212,52 @@ sub pb_conf_read {
 
 my $conffile = shift;
 my @param = @_;
-my $trace;
 my @ptr;
-my %h;
+my $lh;
 
-open(CONF,$conffile) || die "Unable to open $conffile";
-while(<CONF>) {
-	if (/^\s*([A-z0-9-_.]+)\s+([[A-z0-9-_.]+)\s*=\s*(.+)$/) {
-		pb_log(3,"DEBUG: 1:$1 2:$2 3:$3\n");
-		$h{$1}{$2}=$3;
-	}
-}
-close(CONF);
+$lh = pb_conf_cache($conffile,$lh);
 
-for my $param (@param) {
-	push @ptr,$h{$param};
+foreach my $param (@param) {
+	push @ptr,$lh->{$param};
 }
 return(@ptr);
 }
 
+
+
+=item B<pb_conf_get_in_hash_if>
+
+This function returns a table, corresponding to a set of values queried in the hash passed in parameter or undef if it doesn't exist. 
+It takes a table of keys as an input parameter.
+
+=cut
+
+sub pb_conf_get_in_hash_if {
+
+my $lh = shift || return(());
+my @params = @_;
+my @ptr = ();
+
+pb_log(2,"DEBUG: pb_conf_get_in_hash_if on params ".join(' ',@params)."\n");
+foreach my $k (@params) {
+	push @ptr,$lh->{$k};
+}
+
+pb_log(2,"DEBUG: pb_conf_get_in_hash_if returns\n".Dumper(@ptr));
+return(@ptr);
+}
+
+
+
 =item B<pb_conf_get_if>
 
-This function returns a table, corresponding to a set of values queried in the conf files or undef if it doen't exist. It takes a table of keys as an input parameter.
+This function returns a table, corresponding to a set of values queried in the %h hash or undef if it doen't exist. It takes a table of keys as an input parameter.
 
 The format of the configurations file is as follows:
 
 key tag = value1,value2,...
 
-It will gather the values from all the configurations files passed to pb_conf_add, and return the values for the keys, taking in account the order of conf files, to manage overloading.
+It will gather the values from all the configurations files passed to pb_conf_add, and return the values for the keys
 
   $ cat $HOME/.pbrc
   pbver pb = 1
@@ -200,65 +282,49 @@ Valid chars for keys and tags are letters, numbers, '-' and '_'.
 
 sub pb_conf_get_if {
 
-my @param = @_;
-
-my $ptr = undef;
-
-# the most important conf file is first, so read them in reverse order
-foreach my $f (reverse sort { $pbconffiles{$a} <=> $pbconffiles{$b} } keys %pbconffiles) {
-	pb_log(2,"DEBUG: pb_conf_get_if in file $f\n");
-	$ptr = pb_conf_get_fromfile_if("$f",$ptr,@param);
+return(pb_conf_get_in_hash_if($h,@_));
 }
 
-return(@$ptr);
-}
+=item B<pb_conf_add_last_in_hash>
 
-=item B<pb_conf_fromfile_if>
+This function merges the values passed in the hash parameter into the %h hash, but only if itdoesn't already contain a value, or if the value is more precise (real value instead of default)
 
-This function returns a pointer on a table, corresponding to a merge of values queried in the conf file and the pointer on another table passed as parameter. It takes a table of keys as last input parameter.
-
-  my ($k1) = pb_conf_fromfile_if("$HOME/.pbrc",undef,"pbver","pblist");
-  my ($k2) = pb_conf_fromfile_if("$HOME/.pbrc3",$k1,"pbver","pblist");
-
-It is used internally by pb_conf_get_if and is not exported yet.
+It is used internally by pb_conf_add and is not exported.
 
 =cut
 
+sub pb_conf_add_last_in_hash {
 
-sub pb_conf_get_fromfile_if {
+my $ptr = shift || undef;
 
-my $conffile = shift;
-my $ptr2 = shift || undef;
-my @param = @_;
+return if (not defined $ptr);
+# TODO: test $ptr is a hash pointer
 
-# Everything is returned via ptr1
-my @ptr1 = ();
-my @ptr2 = ();
+# When called without correct initialization, try to work anyway with default as project
+pb_conf_init("default") if (not defined $ENV{'PBPROJ'});
 
-# @ptr1 contains the values overloading what @ptr2 may contain.
-@ptr1 = pb_conf_read_if("$conffile", @param) if (defined $conffile);
-@ptr2 = @$ptr2 if (defined $ptr2);
+my @params = (sort keys %$ptr);
+
+# Everything is returned via @h
+# @h contains the values overloading what @ptr may contain.
+my @h = pb_conf_get_if(@params);
+my @ptr = pb_conf_get_in_hash_if($ptr,@params);
 
 my $p1;
 my $p2;
 
-pb_log(2,"DEBUG: pb_conf_get_from_file $conffile: ".Dumper(@ptr1)."\n");
-pb_log(2,"DEBUG: pb_conf_get_from_file input: ".Dumper(@ptr2)."\n");
-pb_log(2,"DEBUG: pb_conf_get_from_file param: ".Dumper(@param)."\n");
+pb_log(2,"DEBUG: pb_conf_add_last_in_hash params: ".Dumper(@params)."\n");
+pb_log(2,"DEBUG: pb_conf_add_last_in_hash hash: ".Dumper(@h)."\n");
+pb_log(2,"DEBUG: pb_conf_add_last_in_hash input: ".Dumper(@ptr)."\n");
 
-foreach my $i (0..$#param) {
-	$p1 = $ptr1[$i];
-	# Optimisation doesn't seem useful
-	# if ((defined $p1) && (defined $cachedval{$p1})) {
-	# $ptr1[$i] = $cachedval{$p1};
-	# next;
-	# }
-	$p2 = $ptr2[$i];
-	# Always try to take the param from ptr1 
-	# in order to mask what could be defined already in ptr2
+foreach my $i (0..$#params) {
+	$p1 = $h[$i];
+	$p2 = $ptr[$i];
+	# Always try to take the param from h 
+	# in order to mask what could be defined already in ptr
 	if (not defined $p2) {
 		# exit if no p1 either
-		next if ((not defined $p1) || (not defined $ENV{'PBPROJ'}));
+		next if (not defined $p1);
 		# No ref in p2 so use p1
 		$p1->{$ENV{'PBPROJ'}} = $p1->{'default'} if ((not defined $p1->{$ENV{'PBPROJ'}}) && (defined $p1->{'default'}));
 	} else {
@@ -277,7 +343,7 @@ foreach my $i (0..$#param) {
 
 			if (not defined $p1->{$ENV{'PBPROJ'}}) {
 				if (defined $p2->{$ENV{'PBPROJ'}}) {
-					$p1->{$ENV{'PBPROJ'}} = $p2->{$ENV{'PBPROJ'}} if (defined $p2->{$ENV{'PBPROJ'}});
+					$p1->{$ENV{'PBPROJ'}} = $p2->{$ENV{'PBPROJ'}};
 				} else {
 					$p1->{$ENV{'PBPROJ'}} = $p1->{'default'} if (defined $p1->{'default'});
 				}
@@ -289,12 +355,9 @@ foreach my $i (0..$#param) {
 			}
 		}
 	}
-	$ptr1[$i] = $p1;
-	# Cache values to avoid redoing all that analyze when asked again on a known value
-	# $cachedval{$p1} = $p1;
+	$h->{$params[$i]} = $p1;
 }
-pb_log(2,"DEBUG: pb_conf_get output: ".Dumper(@ptr1)."\n");
-return(\@ptr1);
+pb_log(2,"DEBUG: pb_conf_add_last_in_hash output: ".Dumper($h)."\n");
 }
 
 =item B<pb_conf_get>
@@ -315,14 +378,41 @@ if (not defined $ENV{'PBPROJ'}) {
 	$proj = $ENV{'PBPROJ'};
 }
 
-die "No params found for $proj" if (not @return);
+confess "No params found for $proj" if (not @return);
 
 foreach my $i (0..$#param) {
-	die "No $param[$i] defined for $proj" if (not defined $return[$i]);
+	confess "No $param[$i] defined for $proj" if (not defined $return[$i]);
 }
 return(@return);
 }
 
+
+=item B<pb_conf_get_all>
+
+This function returns an array wirh all configuration parameters
+
+=cut
+
+sub pb_conf_get_all {
+
+return(sort keys %$h);
+}
+
+
+=item B<pb_conf_print>
+
+This function prints every configuration parameter in order to help debug stacking issues with conf files
+
+=cut
+
+sub pb_conf_print {
+
+pb_log(0,"Full pb configuration for project $ENV{'PBPROJ'}\n");
+pb_log(0,"====================================\n");
+foreach my $k (pb_conf_get_all()) {
+	pb_log(0,"$k => ".Dumper($h->{$k})."\n");
+}
+}
 =back 
 
 =head1 WEB SITES
